from .MSODrawingGroup import OfficeArtDggContainer
import os, struct

def XlsGetImages(path:str) -> list:
    """
    获得 .xls 文件的图片数据.

    Args:
        path (str): 输入文件目录

    Return:
        list[tuple]: 返回以 tuple(图片后缀, bytes) 组成的列表
    """
    ret = []
    if not os.path.exists(path):
        return ret
    with open(path, "rb") as f:
        f.seek(0, os.SEEK_END)
        size = f.tell()
        f.seek(0, os.SEEK_SET)
        xlsBin = f.read(size)
    # 文件格式标识
    identifier = xlsBin[:8]
    if b"\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1" != identifier:
        # 文件格式标识不正确
        return ret
    # 段大小
    secSize = 2 ** struct.unpack("<H", xlsBin[30: 32])[0]
    # 短段大小
    shortSecSize = 2 ** struct.unpack("<H", xlsBin[32:34])[0]
    # SAT 用的段数量
    secAllocNumUsed = struct.unpack("<I", xlsBin[44:48])[0]
    # 目录段ID
    directorySecID = struct.unpack("<i", xlsBin[48:52])[0]
    # 标准流最小大小
    minimumSizeOfStandardStream = struct.unpack("<I", xlsBin[56: 60])[0]
    # SSAT 段ID
    shortAllocTableSecID = struct.unpack("<i", xlsBin[60: 64])[0]
    # SSAT 占用的段数量
    shortSecNumUsed = struct.unpack("<I", xlsBin[64: 68])[0]
    # MSAT 段ID
    masterAllocTableSecID = struct.unpack("<i", xlsBin[68:72])[0]
    # MSAT 占用的段数量
    masterSecNumUsed = struct.unpack("<I", xlsBin[72: 76])[0]

    def get_sec_off(sec_id):
        """获得段偏移"""
        return 512 + sec_id * secSize
    def get_short_sec_off(sec_id):
        """获得短段偏移"""
        return short_id * shortSecSize
    def to_alloc_table(sec_id):
        """转成分配表"""
        t = xlsBin[get_sec_off(sec_id):get_sec_off(sec_id) + secSize]
        return [struct.unpack("<i", t[i : i + 4])[0]
                for i in range(0, len(t), 4)]
    def get_data(sec_id, table):
        """从分配表获取数据"""
        chain = [sec_id]
        i = sec_id
        ret = bytearray()
        while (i := table[i]) >= 0:
            chain.append(i)
        for i in chain:
            off = get_sec_off(i)
            ret.extend(xlsBin[off:off + secSize])
        return ret
    def get_sec_data(sec_id):
        """获得段数据"""
        off = get_sec_off(sec_id)
        return xlsBin[off:off + secSize]
    # MSAT
    masterAllocTable = bytearray()
    masterAllocTable.extend(xlsBin[76: 76 + 436])
    masterAllocTable = [struct.unpack("<i", masterAllocTable[i: i + 4])[0]
                        for i in range(0, len(masterAllocTable), 4)]
    # SAT, 接下来拼接 SAT表
    secAllocTable = []
    for i in masterAllocTable:
        # MSAT 存 SAT, 且是顺序存
        if i >= 0:
            secAllocTable.extend(to_alloc_table(i))
            continue
        break
    # 完善 MSAT
    while True:
        if masterAllocTableSecID >= 0:
            msatNext = to_alloc_table(masterAllocTableSecID)
            masterAllocTableSecID = msatNext.pop(len(msatNext) - 1)
            # 完善 SAT
            for i in msatNext:
                if i >= 0:
                    secAllocTable.extend(to_alloc_table(i))
                    continue
                break
            masterAllocTable.extend(msatNext)
            del msatNext
            continue
        break
    # 获得目录
    directory = get_data(directorySecID, secAllocTable)
    # 获得目录项
    directoryEntries = [struct.unpack("<64sHBBiii16siQQiIi",
                                      directory[i: i + 128])
                        for i in range(0, len(directory), 128)]
    del directory
    directoryFields = ["name", "name_size", "type", "node_color",
                       "left_child_node_id", "right_child_node_id",
                       "root_node_entry_id", "unique_identifier",
                       "user_flag", "timestamp_creation",
                       "timestamp_modification", "short_sec_or_sec_id",
                       "total_stream_size", "not_used"]
    directoryEntries = [dict(zip(directoryFields, i))
                        for i in directoryEntries]
    # 寻找 Workbook
    for i in directoryEntries:
        i["name"]:str = i["name"].decode(encoding="utf-16-le") \
            .rstrip(b'\x00'.decode())
        if i["name"] == "Workbook":
            # 获取用户流
            stream = get_data(i["short_sec_or_sec_id"], secAllocTable)
    # 不再需要 xlsBin
    del xlsBin
    # 获得内存视图
    streammv = memoryview(stream)
    # 所有 BIFF 对象
    BIFFs = []
    curPos = 0
    # 获得流的所有 BIFF 对象
    while True:
        if curPos + 4 < streammv.nbytes:
            biffSize = struct.unpack("<H", streammv[curPos + 2:curPos + 4])[0]
            biffSize = biffSize + 4
            BIFFs.append((curPos, biffSize, streammv[curPos:curPos + 2],
                          streammv[curPos + 4:curPos + biffSize]))
            curPos += biffSize
            continue
        break
    # 获得 Microsoft Office 绘图组 (MSODrawingGroup)
    curPos = 0
    drawingGroup = []
    tBIFF = bytearray()
    for i in BIFFs:
        if curPos == 0:
            # 判断绘图组标识和继续标识
            if i[2] == b'\xeb\x00' or i[2] == b'\x3c\x00':
                curPos = i[0] + i[1]
                tBIFF.extend(i[3])
                continue
        else:
            if curPos != i[0]:
                drawingGroup.append(tBIFF)
                tBIFF.clear()
                curPos = 0
                continue
            tBIFF.extend(i[3])
            curPos += i[1]
    if len(tBIFF) > 0:
        drawingGroup.append(tBIFF)
    del tBIFF
    del curPos
    # 判断是否存在绘图组
    if len(drawingGroup) == 0:
        return ret
    # 获得美术容器
    try:
        artContianer = OfficeArtDggContainer(drawingGroup[0])
        for blip in artContianer.blipStore.rgfb:
            ret.append((blip._blip.suffix, blip._blip.BLIPFileData))
    except:
        pass
    return ret



"""
创建: 2022年9月15日
"""
from struct import unpack


class OfficeArtRecordHeader:
    def __init__(self, data):
        tmv = memoryview(data)
        tVal = unpack("<H", tmv[0:2])[0]
        self.recVer = tVal & 0xf
        self.recInstance = tVal >> 4
        tVal = unpack("<H", tmv[2:4])[0]
        self.recType = tVal
        self.recLen = unpack("<I", tmv[4:8])[0]

    def GetTotalSize(self):
        return self.recLen + 8


class OfficeArtFDGG:
    def __init__(self, data):
        tmv = memoryview(data)
        self.spidMax = unpack("<I", tmv[0:4])[0]
        self.cidcl = unpack("<I", tmv[4:8])[0]
        self.cspSaved = unpack("<I", tmv[8:12])[0]
        self.cdgSaved = unpack("<I", tmv[12:16])[0]


class OfficeArtIDCL:
    def __init__(self, data):
        tmv = memoryview(data)
        self.dgid = unpack("<I", tmv[0:4])[0]
        self.cspidCur = unpack("<I", tmv[4:8])[0]


class OfficeArtFDGGBlock:
    def __init__(self, data):
        tmv = memoryview(data)
        self.rh = OfficeArtRecordHeader(tmv[0:8])
        self.head = OfficeArtFDGG(tmv[8:24])
        self.Rgidcl = [OfficeArtIDCL(tmv[24 + 8 * i: 24 + 8 * i + 8])
                       for i in range(self.head.cidcl - 1)]

class OfficeArtBlip:
    OfficeArtBlipEMFType = 0xF01A # as defined in section 2.2.24.
    OfficeArtBlipWMFType = 0xF01B # as defined in section 2.2.25.
    OfficeArtBlipPICTType = 0xF01C # as defined in section 2.2.26.
    OfficeArtBlipJPEGType = 0xF01D # as defined in section 2.2.27.
    OfficeArtBlipPNGType = 0xF01E # as defined in section 2.2.28.
    OfficeArtBlipDIBType = 0xF01F # as defined in section 2.2.29.
    OfficeArtBlipTIFFType = 0xF029 # as defined in section 2.2.30.
    OfficeArtBlipJPEGType2 = 0xF02A # 0xF02A is treated as 0xF01D.

    class MSOBLIPTYPE:
        msoblipERROR = 0x00 # Error reading the file.
        msoblipUNKNOWN = 0x01 # Unknown BLIP type.
        msoblipEMF = 0x02 # EMF.
        msoblipWMF = 0x03 # WMF.
        msoblipPICT = 0x04 # Macintosh PICT.
        msoblipJPEG = 0x05 # JPEG.
        msoblipPNG = 0x06 # PNG.
        msoblipDIB = 0x07 # DIB
        msoblipTIFF = 0x11 # TIFF
        msoblipCMYKJPEG = 0x12 # JPEG in the YCCK or CMYK color space.

    class OfficeArtBlipJPEG:
        suffix = "jpeg"
        def __init__(self, rh:OfficeArtRecordHeader, data):
            tmv = memoryview(data)
            self.rgbUid1 = tmv[0:16]
            tpos = 16
            if rh.recInstance == 0x46b or \
                rh.recInstance == 0x6e3:
                self.rgbUid2 = tmv[tpos:tpos + 16]
                tpos += 16
            self.tag = unpack("B", tmv[tpos:tpos + 1])
            tpos += 1
            self.BLIPFileDataSize = rh.recLen - tpos
            self.BLIPFileData = tmv[tpos:tpos + self.BLIPFileDataSize]

    class OfficeArtBlipPNG:
        suffix = "png"
        def __init__(self, rh:OfficeArtRecordHeader, data):
            tmv = memoryview(data)
            self.rgbUid1 = tmv[0:16]
            tpos = 16
            if rh.recInstance == 0x6e1:
                self.rgbUid2 = tmv[tpos:tpos + 16]
                tpos += 16
            self.tag = unpack("B", tmv[tpos:tpos + 1])
            tpos += 1
            self.BLIPFileDataSize = rh.recLen - tpos
            self.BLIPFileData = tmv[tpos:tpos + self.BLIPFileDataSize]

    def __init__(self, data):
        tmv = memoryview(data)
        self.rh = OfficeArtRecordHeader(tmv[0:8])
        if self.rh.recType == self.OfficeArtBlipJPEGType or \
            self.rh.recType == self.OfficeArtBlipJPEGType2:
            self._blip = self.OfficeArtBlipJPEG(self.rh,
                                                tmv[8:self.rh.GetTotalSize()])
        elif self.rh.recType == self.OfficeArtBlipPNGType:
            self._blip = self.OfficeArtBlipPNG(self.rh,
                                               tmv[8:self.rh.GetTotalSize()])


class OfficeArtFBSE:
    def __init__(self, data):
        tmv = memoryview(data)
        self.rh = OfficeArtRecordHeader(tmv[0:8])
        self.btWin32 = unpack("<B", tmv[8:9])[0]
        self.btMacOS = unpack("<B", tmv[9:10])[0]
        self.rgbUid = tmv[10:26]
        self.tag = tmv[26:28]
        self.size = unpack("<I", tmv[28:32])[0]
        self.cRef = unpack("<I", tmv[32:36])[0]
        self.foDelay = tmv[36:40]
        self.cbName = unpack("<B", tmv[41:42])[0]
        self.nameData = None
        tpos = 44
        if self.cbName > 0:
            self.nameData = unpack("{}s".format(self.cbName),
                                   tmv[tpos:tpos + self.cbName])[0]
            self.nameData:str = self.nameData.decode("utf-16-le")
            self.nameData.rstrip(b"\x00".decode())
            tpos += self.cbName
        self.embedded = OfficeArtBlip(tmv[tpos:tpos + self.rh.recLen])
        self._blip = self.embedded._blip

def OfficeArtBStoreContainerFileBlock(data):
    tmv = memoryview(data)
    theader = OfficeArtRecordHeader(tmv[0:8])
    if theader.recType == 0xF007:
        return OfficeArtFBSE(tmv[0:theader.GetTotalSize()])
    return OfficeArtBlip(tmv[0:theader.GetTotalSize()])

class OfficeArtBStoreContainer:
    def __init__(self, data):
        tmv = memoryview(data)
        self.rh = OfficeArtRecordHeader(tmv[0:8])
        self.rgfb = []
        tpos = 8
        for i in range(self.rh.recInstance):
            theader = OfficeArtRecordHeader(tmv[tpos:tpos + 8])
            tstore = OfficeArtBStoreContainerFileBlock(
                tmv[tpos:tpos + theader.GetTotalSize()])
            tpos += theader.GetTotalSize()
            self.rgfb.append(tstore)


###########################################################################

class OfficeArtDggContainer:
    def __init__(self, data: bytes):
        self.dataMv = memoryview(data)
        tmv = self.dataMv
        tpos = 0
        self.rh = OfficeArtRecordHeader(tmv[tpos:tpos + 8])
        tpos += 8
        theader = OfficeArtRecordHeader(tmv[tpos:tpos + 8])
        self.drawingGroup = OfficeArtFDGGBlock(
            tmv[tpos:tpos + theader.GetTotalSize()])
        tpos += theader.GetTotalSize()
        theader = OfficeArtRecordHeader(tmv[tpos:tpos + 8])
        self.blipStore = OfficeArtBStoreContainer(
            tmv[tpos:tpos + theader.GetTotalSize()])
        # self.drawingPrimaryOptions
        # self.drawingTertiaryOptions
        # self.colorMRU
        # self.splitColors

    def __del__(self):
        if type(self.dataMv) == memoryview:
            self.dataMv.release()



from XlsImage import *

if __name__ == "__main__":
    import sys, os
    allFuncs = dir()
    allObjs = vars()
    argv = sys.argv
    Usage = f'''usage: XlsImage <xls file> [output dir]
'''
    if len(argv) > 1:
        argv = argv[1:]
        output = ""
        xlsFile = argv.pop(0)
        if argv:
            output = os.path.join(output, argv.pop(0))
        try:
            os.makedirs(output)
        except:
            pass
        if not os.path.exists(xlsFile):
            print("file doesn't exsits!")
            exit(-1)
        imgs = XlsGetImages(xlsFile)
        print("total images: ", len(imgs))
        for k, v in enumerate(imgs):
            filename = "{}.{}".format(k, v[0])
            filename = os.path.join(output, filename)
            print("write to: ", filename)
            with open(filename, "wb") as f:
                f.write(v[1])
        print("done!")
    else:
        print(Usage)


import src.XlsImage as xls

def xls_test():
    imgs = xls.XlsGetImages("./untrack/test.xls")
    print("total images: ", len(imgs))
    for k, v in enumerate(imgs):
        with open("./untrack/{}.{}".format(k, v[0]), "wb") as f:
            f.write(v[1])



if __name__ == "__main__":
    import sys
    allFuncs = dir()
    allObjs = vars()
    argv = sys.argv
    if len(argv) > 1:
        funcName = argv[1]
        argv = argv[1:]
        if funcName == "-a":
            for i in allFuncs:
                if callable(allObjs[i]):
                    print("----- {} -----".format(i))
                    allObjs[i]()
        else:
            for i in argv:
                if i in allFuncs:
                    print("----- {} -----".format(i))
                    allObjs[i]()
                else:
                    print("没有函数: " + funcName)
    else:
        print("使用方法:\npython mtest.py {测试函数1} {测试函数2} ...\n"
              "python mtest.py -a 执行全测试")